#include <glib.h>
#include <sys/time.h>
#include <unistd.h>

#include "config.h"
#include "netstat.h"
#include "wired.h"
#include "support.h"
#include "node.h"
#include "netobject.h"


extern int gPaused;
extern float gLinkTimeoutTime;


NetStat* NetStatCreate( ip* addr )
{	NetStat* ns = (NetStat*) malloc( sizeof( NetStat ) );

	memcpy( ns->addr, addr, 4*sizeof(ip) );
	ns->receivedFrom = 0;
	ns->sentTo = 0;
	ns->totalNumPackets = 0;
	ns->totalData = 0;
	ns->totalDataNew = 0;
	ns->snapTraffic = 0;
	return ns;
}


void NetStatDestroy( NetStat* ns )
{	if( ns->receivedFrom )
	{	g_tree_traverse( ns->receivedFrom, (GTraverseFunc) NetInfoDestroy, G_IN_ORDER, 0 );
		g_tree_destroy( ns->receivedFrom );
	}
	ns->receivedFrom = 0;
	if( ns->sentTo )
	{	g_tree_traverse( ns->sentTo, (GTraverseFunc) NetInfoDestroy, G_IN_ORDER, 0 );
		g_tree_destroy( ns->sentTo );
	}
	ns->sentTo = 0;
	free( ns );
}


void NetStatForget( NetStat* ns, Node* n )
{	NetInfo* ni = 0;

	if( ns->receivedFrom )
	{	ni = g_tree_lookup( ns->receivedFrom, (gpointer) IPToKey( n->addr ) );
		if( ni )
			g_tree_remove( ns->receivedFrom, (gpointer) IPToKey( n->addr ) );
	}
	if( ns->sentTo )
	{	if( ! ni )
			ni = g_tree_lookup( ns->sentTo, (gpointer) IPToKey( n->addr ) );
		if( ni )
			g_tree_remove( ns->sentTo, (gpointer) IPToKey( n->addr ) );
	}
	if( ni )
	{	DEBUG(2) printf( "Node %s is forgetting node %s\n", IPToString( ns->addr ), IPToString( n->addr ) );
		if( ni->visual )
		{	RemoveLinkVisual( ni );
		}
		NetInfoDestroy( 0, ni, 0 );
	}
}


int NetStatProcess( NetStat* ns, Node* n, GTree* gt, ip* addr,
	struct timeval* timestamp, guint32 data, char isSent )
{	NetInfo* ni;
	char temp[20];

	if( ! (ni = FindNetInfo( gt, addr ) ) )
	{	ni = CreateNetInfo( addr, n, isSent );
		g_tree_insert( gt, (gpointer) IPToKey( ni->addr ), (gpointer) ni );
	}

	ni->numPackets++;
	ni->totalDataNew += data;
	ni->lastPacket.tv_sec = timestamp->tv_sec;
	ni->lastPacket.tv_usec = timestamp->tv_usec;

	ns->totalNumPackets++;
	ns->totalDataNew += data;
	ns->lastPacket.tv_sec = timestamp->tv_sec;
	ns->lastPacket.tv_usec = timestamp->tv_usec;
	return 0;
}


int NetStatProcessSend( NetStat* ns, Node* n, ip* dest, struct timeval* timestamp,
	guint32 data )
{	if( ! ns->sentTo )
		ns->sentTo = g_tree_new( KeyCompare );
	return NetStatProcess( ns, n, ns->sentTo, dest, timestamp, data, 1 );
}


int NetStatProcessReceive( NetStat* ns, Node* n, ip* source, struct timeval* timestamp,
	guint32 data )
{	if( ! ns->receivedFrom )
		ns->receivedFrom = g_tree_new( KeyCompare );
	return NetStatProcess( ns, n, ns->receivedFrom, source, timestamp, data, 0 );
}


int NetStatDisplayInfo( NetStat* ns )
{	printf( "\t\treceived:\n" );
	if( ns->receivedFrom )
		g_tree_traverse( ns->receivedFrom, &NetInfoDisplayInfo, G_IN_ORDER, 0 );
	printf( "\t\tsent:\n" );
	if( ns->sentTo )
		g_tree_traverse( ns->sentTo, &NetInfoDisplayInfo, G_IN_ORDER, 0 );
	printf( "\t\taverage traffic: %lu\n", GetAverageTraffic( ns->totalNumPackets ) );
	return 0;
}


void NetStatUpdate( NetStat* ns, struct timeval* current )
{	char temp[20];
	float t;
//	ns->snapTraffic = GetSnapshotTraffic( &ns->lastPacket, current );
	ns->snapTraffic = GetAverageTraffic( ns->totalData );
	ns->totalNumPackets = 0;
	ns->totalData = ns->totalDataNew;
	ns->totalDataNew = 0;

	if( ns->sentTo )
		g_tree_traverse( ns->sentTo, (GTraverseFunc) NetInfoUpdate,
			G_IN_ORDER, current );
	if( ns->receivedFrom )
		g_tree_traverse( ns->receivedFrom, (GTraverseFunc) NetInfoUpdate,
			G_IN_ORDER, current );
}


NetInfo* FindNetInfo( GTree* infoTree, ip* key )
{	return g_tree_lookup( infoTree, (gpointer) IPToKey( key ) );
}


NetInfo* CreateNetInfo( ip* addr, Node* n, char isSent )
{	NetInfo* ni = (NetInfo*) malloc( sizeof( NetInfo ) );

	memcpy( ni->addr, addr, 4*(sizeof( ip )) );
	ni->numPackets = 0;
	ni->totalData = 0;
	ni->totalDataNew = 0;
	ni->snapTraffic = 0;
	ni->myself = n;
	ni->isSent = isSent;
	ni->nodeLink = FindNode( addr );
	DEBUG(1) printf( "Creating new netinfo for node %s to node %s\n", IPToString( n->addr), IPToString( addr ) );
	ni->visual = AddLinkVisual( ni );
	return ni;
}


int NetInfoDestroy( gpointer unused1, NetInfo* ni, gpointer unused2 )
{	if( ni->visual )
	{	DEBUG(2) printf( "calling remove in netinfodestroy, visual is %p\n", ni->visual );
		RemoveLinkVisual( ni );
	}
	free( ni );
	return 0;
}


int NetInfoForget( gpointer unused, NetInfo* ni, Node* f )
{	if( ni->nodeLink == f )
	{	if( ni->visual )
		{	printf( "calling remove in netinfoforget, visual is %p\n", ni->visual );
			RemoveLinkVisual( ni );
		}
		ni->nodeLink = 0;
	}
	return 0;
}


int NetInfoDisplayInfo( gpointer dummy1, gpointer g, gpointer dummy2 )
{	NetInfo* ni = (NetInfo*) g;

	printf( "\t\t\t%s: %ld\n", IPToString( ni->addr ), ni->numPackets );
	printf( "\t\t\taverage traffic: %lu\n", GetAverageTraffic( ni->numPackets ) );
	return 0;
}


int NetInfoUpdate( gpointer unused, NetInfo* ni, struct timeval* current )
{	float t;
	ni->snapTraffic = GetAverageTraffic( ni->totalData );
	ni->numPackets = 0;
	ni->totalData = ni->totalDataNew;
	ni->totalDataNew = 0;
	if( ! ni->nodeLink )
		ni->nodeLink = FindNode( ni->addr );
	t = GetTimeDifference2( &ni->lastPacket, current );

	if( ! gPaused )
	{	if( ni->visual && t > gLinkTimeoutTime )
		{	DEBUG(2) printf( "calling remove from netinfoupdate, visual is %p\n", ni->visual );
			RemoveLinkVisual( ni );
		}
		else if( ! ni->visual && t < gLinkTimeoutTime )
		{	DEBUG(2) printf( "Re-creating new netinfo\n" );
			AddLinkVisual( ni );
		}
	}

	return 0;
}


guint32 NetInfoKey( NetInfo* ni )
{       return IPToKey( ni->addr );
}


gint NetInfoCompare( gconstpointer k1, gconstpointer k2 )
{       return (guint32) k1 - (guint32) k2;
}
